@@ -8,8 +8,12 @@ use Carp;
 use File::Spec;
 
 use Net::SFTP::Foreign::Attributes;
-use Net::SFTP::Foreign::Constants qw(:error);
-use Net::SFTP::Foreign::Helpers qw(_sort_entries _gen_wanted _do_nothing);
+use Net::SFTP::Foreign::Constants qw( :error );
+
+use Net::SFTP::Foreign::Helpers qw( _sort_entries
+				    _gen_wanted
+				    _do_nothing );
+
 use Net::SFTP::Foreign::Common;
 our @ISA = qw(Net::SFTP::Foreign::Common);
 
@@ -29,7 +33,7 @@ sub stat {
     $! = 0;
     my $a = Net::SFTP::Foreign::Attributes->new_from_stat(CORE::stat($_[1]));
     unless ($a) {
-	$_[0]->_set_error(SFTP_ERR_LOCAL_STAT_FAILED, "Couldn't stat local file '$_[1]'", $!);
+	$_[0]->_set_error(SFTP_ERR_LOCAL_STAT_FAILED, "Couldn't stat local file '$_[1]': $!");
     }
     $a
 }
@@ -38,7 +42,7 @@ sub lstat {
     $! = 0;
     my $a = Net::SFTP::Foreign::Attributes->new_from_stat(CORE::lstat($_[1]));
     unless ($a) {
-	$_[0]->_set_error(SFTP_ERR_LOCAL_STAT_FAILED, "Couldn't stat local file '$_[1]'", $!);
+	$_[0]->_set_error(SFTP_ERR_LOCAL_STAT_FAILED, "Couldn't stat local file '$_[1]': $!");
     }
     $a
 }
@@ -47,7 +51,7 @@ sub readlink {
     $! = 0;
     my $target = readlink $_[1];
     unless (defined $target) {
-	$_[0]->_set_error(SFTP_ERR_LOCAL_READLINK_FAILED, "Couldn't read link '$_[1]'", $!);
+	$_[0]->_set_error(SFTP_ERR_LOCAL_READLINK_FAILED, "Couldn't read link '$_[1]': $!");
     }
     $target
 }
@@ -55,8 +59,7 @@ sub readlink {
 sub join {
     shift;
     my $path = File::Spec->join(@_);
-    $path = File::Spec->canonpath($path);
-    print 'lfs->join("'.join('", "', @_)."\") => $path\n";
+    1 while ($path =~ s/([\/\\])\.(?:[\/\\]|$)/$1/g);
     $path
 }
 
